# pylint: disable=import-outside-toplevel
import logging
import math
import time

import orjson

from seagoat import __version__
from seagoat.queue.base_queue import BaseQueue
from seagoat.queue.base_queue import LOW_PRIORITY


SECONDS_BETWEEN_MAINTENANCE = 10


def calculate_accuracy(chunks_analyzed: int, total_chunks: int) -> int:
    if total_chunks == 0 or total_chunks - chunks_analyzed == 0:
        return 100

    progress = chunks_analyzed / total_chunks

    k = 10
    x_0 = 0.25

    f_x = 1 / (1 + math.exp(-k * (progress - x_0)))
    f_0 = 1 / (1 + math.exp(k * x_0))
    f_1 = 1 / (1 + math.exp(-k * (1 - x_0)))

    normalized_value = (f_x - f_0) / (f_1 - f_0)
    percentage_value = normalized_value * 100
    rounded_percentage_value = int(percentage_value)

    if percentage_value > 0 and rounded_percentage_value == 0:
        return 1

    return rounded_percentage_value


class TaskQueue(BaseQueue):
    def _get_context(self):
        context = super()._get_context()

        from seagoat.engine import Engine

        seagoat_engine = Engine(self.kwargs["repo_path"])
        context["seagoat_engine"] = seagoat_engine
        context["last_maintenance"] = None
        context["last_repo_state_hash"] = None
        return context

    def handle_maintenance(self, context):
        if (
            context["last_maintenance"] is not None
            and time.time() - context["last_maintenance"] < SECONDS_BETWEEN_MAINTENANCE
        ):
            return

        current_repo_state_hash = context["seagoat_engine"].repository.get_status_hash()

        # Do not re-analyze repo if nothing changed
        if context["last_repo_state_hash"] == current_repo_state_hash:
            return

        context["last_repo_state_hash"] = current_repo_state_hash
        context["last_maintenance"] = time.time()

        if self._task_queue.qsize() > 0:
            return

        logging.info("Checking repository for new changes")
        remaining_chunks_to_analyze = context["seagoat_engine"].analyze_codebase(
            self.kwargs["minimum_chunks_to_analyze"]
        )

        logging.info("Analyzed the minimum number of chunks needed to operate. ")
        if remaining_chunks_to_analyze:
            logging.info(
                "Note, %s chunks need to be analyzed for optimum performance.",
                len(remaining_chunks_to_analyze),
            )

            for chunk in remaining_chunks_to_analyze:
                self.enqueue(
                    "analyze_chunk", chunk, priority=LOW_PRIORITY, wait_for_result=False
                )
        else:
            logging.info("Analyzed all chunks!")

    def handle_analyze_chunk(self, context, chunk):
        logging.info("Note, %s tasks left in the queue.", self._task_queue.qsize())
        logging.info("Processing chunk %s...", chunk)
        context["seagoat_engine"].process_chunk(chunk)

        if self._task_queue.qsize() == 0:
            logging.info("Analyzed all chunks!")

    def handle_query(self, context, **kwargs):
        context["seagoat_engine"].query(kwargs["query"])
        context["seagoat_engine"].fetch_sync(
            limit_clue=kwargs["limit_clue"],
            context_above=int(kwargs["context_above"]),
            context_below=int(kwargs["context_below"]),
        )
        results = context["seagoat_engine"].get_results(kwargs["limit_clue"])
        formatted_results = [result.to_json(kwargs["query"]) for result in results]

        serialized_results = orjson.dumps(
            {
                "results": formatted_results,
                "version": __version__,
            }
        )

        return serialized_results

    def handle_get_stats(self, context):
        engine = context["seagoat_engine"]
        analyzed_count = len(engine.cache.data["chunks_already_analyzed"])
        unanalyzed_count = len(engine.cache.data["chunks_not_yet_analyzed"])
        total_chunks = analyzed_count + unanalyzed_count

        return {
            "queue": {
                "size": self._task_queue.qsize(),
            },
            "chunks": {
                "analyzed": analyzed_count,
                "unanalyzed": unanalyzed_count,
            },
            "accuracy": {
                "percentage": calculate_accuracy(analyzed_count, total_chunks),
            },
        }



import logging
import mmap
import platform
import re
import subprocess
import tempfile
from pathlib import Path

from seagoat.cache import Cache
from seagoat.repository import Repository
from seagoat.result import Result
from seagoat.sources.chroma import MAXIMUM_VECTOR_DISTANCE
from seagoat.utils.file_reader import read_file_with_correct_encoding
from seagoat.utils.file_types import is_file_type_supported

KILOBYTE = 1024
MEGABYTE = KILOBYTE * 1024
MAX_MMAP_SIZE = 100
MAX_MMAP_SIZE_BYTES = MAX_MMAP_SIZE * MEGABYTE
MAX_FILE_SIZE = 200 * KILOBYTE


class RipGrepCache(str):
    def __init__(self, repository):
        cache = Cache("ripgrep", Path(repository.path), {})
        self.repository = repository
        self.file_path = cache.get_cache_folder() / "mmap"
        if platform.system() == "Windows":
            self.file_path = tempfile.mktemp()
        self.is_initialized = False
        self._data = ""

    def _iterate_files_to_cache(self):
        for file, _ in self.repository.top_files():
            yield file

    def _iterate_lines_to_cache(self):
        for file in self._iterate_files_to_cache():
            file_contents = read_file_with_correct_encoding(file.absolute_path)

            if len(file_contents) > MAX_FILE_SIZE:
                logging.warning("Warning: file %s is too large to cache", file.path)
                continue

            for line_number, line in enumerate(file_contents.splitlines(), start=1):
                yield file, line_number, line

    def _generate_cache_lines(self):
        for file, line_number, line in self._iterate_lines_to_cache():
            yield f"{file.path}:{line_number}:{line}\n"

    def _build_cache_file(self):
        total_estimated_cache_size = 0
        line_count = 0

        with open(self.file_path, "w", encoding="utf-8") as cache_file:
            for formattted_cache_line in self._generate_cache_lines():
                cache_file.write(formattted_cache_line)
                total_estimated_cache_size += len(formattted_cache_line)
                line_count += 1

                if total_estimated_cache_size > MAX_MMAP_SIZE_BYTES:
                    logging.warning(
                        "Warning: maximum estimated ripgrep cache size of %s megabytes exceeded",
                        MAX_MMAP_SIZE,
                    )

                    break

        logging.info(
            "Estimated ripgrep cache size: %.2f megabytes (%s bytes). Line count %s",
            total_estimated_cache_size / MEGABYTE,
            int(total_estimated_cache_size),
            line_count,
        )
        self.is_initialized = True

    def rebuild(self):
        self._build_cache_file()
        if platform.system() == "Windows":
            # Memory map does not work on Windows for some reason
            # Use a simple string as a fallback
            with open(self.file_path, "r", encoding="utf-8") as cache_file:
                self._data = cache_file.read()
        else:
            with open(self.file_path, "r+b") as cache_file:
                self._data = mmap.mmap(cache_file.fileno(), 0)

    # pylint: disable-next=unused-argument
    def encode(self, *args, **kwargs):
        return self._data

    def as_input(self):
        if platform.system() == "Windows":
            return self.encode()

        return self


def _fetch(query_text: str, path: str, limit: int, cache: RipGrepCache):
    query_text = re.sub(r"\s+", "|", query_text)
    files = {}

    cmd = [
        "rg",
        "--max-count",
        str(limit),
        "--ignore-case",
        query_text,
    ]

    try:
        rg_output = subprocess.check_output(
            cmd, encoding="utf-8", input=cache.as_input()
        )
    except subprocess.CalledProcessError as exception:
        rg_output = exception.output
    for line in rg_output.splitlines():
        relative_path, raw_line_number, _ = line.split(":", 2)
        line_number = int(raw_line_number)

        absolute_path = Path(path) / relative_path

        if not is_file_type_supported(relative_path):
            continue

        if relative_path not in files:
            files[relative_path] = Result(str(relative_path), absolute_path)

        # This is so that ripgrep results are on comparable levels with chroma results
        files[relative_path].add_line(line_number, MAXIMUM_VECTOR_DISTANCE * 0.8)

    return files.values()


def initialize(repository: Repository):
    path = repository.path
    memory_cache = RipGrepCache(repository)

    def cache_chunk(_):
        # Ripgrep does not use chunks
        pass

    def cache_repo():
        memory_cache.rebuild()

    def fetch(query_text: str, limit: int):
        if not memory_cache.is_initialized:
            repository.analyze_files()
            cache_repo()

        return _fetch(query_text, str(path), limit, memory_cache)

    return {
        "fetch": fetch,
        "cache_chunk": cache_chunk,
        "cache_repo": cache_repo,
    }
